// BTNCell.cpp: implementation of the BTNCell class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "mfc.h"
#include "BTree.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#include "BTree.h"
#include "Except.h"
#include "Hash.h"
#include <iostream>

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ BTNCell ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
static int hapali = 0;

STRING vectorer( std::vector<BTNCell *> & a);
BTNCell * vetorir( std::vector<BTNCell *> & a, int i);


STRING BTnextToken( STRING input, int &index, char delim)
{
	STRING temp;
	int f=0;
	
	int t = input[index++];
	for(int i = 0; i < t ; i++) 
	{
		temp += input[i+index];	
	}
	index += t;
	return temp;
}

char equal2keys( STRING first, STRING second)
{
	int fs = first.size(), ss = second.size();

	if( ss < 4)
		throw Except("DANGER:Key size cannot be less than 4 bytes!!");

	if( first.size() <= second.size())
		second.resize( first.size());
	else
		second.resize( ss - 4);

	return (first == second);
}

char equalbig2keys( STRING first, STRING second)
{
	int fs = first.size(), ss = second.size();

	if( ss < 4)
		throw Except("DANGER:Key size cannot be less than 4 bytes!!");

	if( first.size() <= second.size())
		second.resize( first.size());
	else
		second.resize( ss - 4);

	return (first >= second);
}

BTNCell::BTNCell( BTree * mypapa, STRING key, BTNode *point, OFFSET byteoffset)
{
	this->reference = NULL; 
	this->mypapa = NULL;

	if( mypapa == NULL)
		throw Except("ERR:NULL pointer passed to BTNodes Constructor!!");

	this->key = key; 
	this->reference = point; 
	this->byteoffset = byteoffset; 
	this->mypapa = mypapa;
	recursive = 1;
}

BTNCell::BTNCell( BTree * mypapa , STRING key, int ObjectID)
{
	this->key = key; 
	reference = NULL; 
	this->ObjectID = ObjectID; 
	this->mypapa = mypapa;
	recursive = 1;
}

BTNCell::~BTNCell()
{
	if( recursive == 1 && reference != NULL)
	{
		delete reference;
		reference = NULL;
	}
}

void BTNCell::merge( BTNCell * first, BTNCell * second)
{
	if( first == NULL || second == NULL)
		throw Except( "ERR:NULL Pointer using!!");

	int temp = first->reference->merge( first->reference, second->reference);
	first->cure();

	if( temp != 0)
	{
		int tempint = mypapa->getplacestring( second->byteoffset).size();
		mypapa->deleteplacestring( second->byteoffset, tempint);

		delete second->reference;
		second->reference = NULL;
	}
	else
		second->cure();
}

void BTNCell::split( BTNCell * first, BTNCell * second)
{
	if( second->key != "" && second->reference != NULL)
		throw Except("ERR:When you break one node of BTree the second one must be empty!!");

	second->reference = new BTNode( first->mypapa, first->reference->flag);
	first->reference->split( first->reference, second->reference);
	first->cure();
	second->cure();
}

void BTNCell::swap( BTNCell *& first, BTNCell *& second)
{
	BTNCell * temp = first;
	first = second;
	second = temp;
}

void BTNCell::cure()
{
	key = reference->bigest();

	STRING temps = reference->pack( mypapa->getfieldterm(), mypapa->getrecordterm());
	STRING temps2 = mypapa->getplacestring( byteoffset);

	if( byteoffset != -1 &&  temps != temps2)
		mypapa->deleteplacestring( byteoffset, temps2.size());

	byteoffset = mypapa->putplacestring( temps);
}

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ BTNode ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
BTNode::BTNode( BTNCell * first, BTNCell * second, BTree * mypapa, char flag)
{
	this->mypapa = NULL;

	if( first == NULL || second == NULL || mypapa == NULL)
		throw Except("ERR:NULL pointer passed to BTNodes Constructor!!");

	mynodes.push_back( first);
	mynodes.push_back( second);
	this->mypapa = mypapa;
	this->flag = flag;
	recursive = 1;

	label = hapali++;
//	std::cout << ">>Node " << label << " constructed." << std::endl;
}

BTNode::BTNode( BTree * mypapa, char flag)
{
	this->mypapa = NULL;

	if( mypapa == NULL)
		throw Except("ERR:NULL pointer passed to BTNodes Constructor!!");

	recursive = 1;
	this->mypapa = mypapa;
	this->flag = flag;

	label = hapali++;
//	std::cout << ">>Node " << label << " constructed." << std::endl;
}

BTNode::~BTNode()
{
	if( recursive == 1 && flag != 1)
	{
		int end = mynodes.size();

		for(int count = 0;count < end;count++)
		{
			delete mynodes[count];
			mynodes[count] = NULL;
		}
	}
//	std::cout << ">>Node " << label << " desstructed." << std::endl;
}

STRING BTNode::bigest()
{ 
	if( mynodes.size() > 0)
		return mynodes[ mynodes.size() - 1]->key;
	else
		return "";
}


char BTNode::add( STRING key, int ObjectID)
{
	int end = mynodes.size() - 1;
	int point;

	for(point = 0;point < end && mynodes[point]->key < key;point++);

	if( flag == 1)
	{
		mynodes.push_back( new BTNCell( mypapa, key, ObjectID));

		for(int anotherend = end + 1;anotherend > 0 && mynodes[anotherend]->key < mynodes[anotherend - 1]->key;anotherend--)
			mynodes[anotherend]->swap( mynodes[anotherend], mynodes[anotherend - 1]);

		if( mynodes.size() > NODESIZE)
			return -1;
		else
			return 0;
	}

	//else
	int tempi = mynodes[point]->reference->add( key, ObjectID);
	mynodes[point]->cure();

	if( tempi == 0)
		return 0;

	//else
	mynodes.push_back( new BTNCell( mypapa));

	for(end++;end - 1> point;end--)
		mynodes[end]->swap( mynodes[end], mynodes[end - 1]);

	if( tempi == -1)
		mynodes[end]->reference = new BTNode( mypapa, 1);
	else if( tempi == 1)
		mynodes[end]->reference = new BTNode( mypapa, 0);
	else 
		throw Except( "ASSERT:tempi-> 0 or 1 or -1!!");

	mynodes[end - 1]->split( mynodes[end - 1], mynodes[end]);

	if( mynodes.size() > NODESIZE)
		return 1;

	//else
	return 0;
}

void BTNode::split( BTNode *first, BTNode *second)
{
	int end = first->mynodes.size();

	for(int count = NODEHALF + 1; count < end;count++)
		second->mynodes.push_back( first->mynodes[count]);

	first->mynodes.resize( NODEHALF + 1);
}

char BTNode::remove( STRING key)
{
	int end = mynodes.size() - 1;
	int point;

	for(point = 0;point < end && mynodes[point]->key < key;point++);

//	if( point == end && mynodes[point]->key > key)
//		throw Except("ERR:Record does not found!!");

	if( flag == 1)
	{
		if( mynodes[point]->key != key)
			throw Except("ERR:Record does not found!!");

		delete mynodes[point];
		mynodes.erase( &mynodes[point]);

		if( mynodes.size() < NODEHALF)
			return 1;
		else
			return 0;
	}

	//else
	int tempi = mynodes[point]->reference->remove( key);
	mynodes[point]->cure();

	if( tempi == 0)
		return 0;

	//else
	if( point == 0)
	{
		mynodes[0]->merge( mynodes[0], mynodes[1]);

		if( mynodes[1]->reference == NULL)
			mynodes.erase( &mynodes[1]);
	}
	else
	{
		mynodes[point - 1]->merge( mynodes[point - 1], mynodes[point]);

		if( mynodes[point]->reference == NULL)
			mynodes.erase( &mynodes[point]);
	}

	if( mynodes.size() < NODEHALF)
		return 1;

	//else
	return 0;
}

int BTNode::merge( BTNode * first, BTNode * second)
{
	if( first == NULL || second == NULL)
		throw Except( "ERR:NULL Pointer using!!");

	int end = second->mynodes.size();

	for(int count = 0;count < end;count++)//Checking for being sorted is not needed
		first->mynodes.push_back( second->mynodes[count]);

	if( first->mynodes.size() < NODESIZE)
		return 1;
	else
	{
		second->mynodes.resize( 0);

		int end = first->mynodes.size();

		for(int count = NODEHALF;count < end;count++)//Checking for being sorted is not needed
			second->mynodes.push_back( first->mynodes[count]);

		first->mynodes.resize( NODEHALF);

		return 0;
	}
}


STRING BTNode::pack( char fterm, char rterm)
{
	STRING temp="";
	temp += fterm;
	temp += flag;
	temp += fterm;
	temp += mynodes.size();
	temp += fterm;

	for(int i=0; i<mynodes.size(); i++)
	{
		temp += mynodes[i]->getKey().size();
		temp += mynodes[i]->getKey();
		temp += fterm;

		if(flag)
		{
			STRING temp1;
			temp1.resize(4);
			int k=0;
			Hash::packInt(temp1,mynodes[i]->ObjectID,k);
//~~~~~
			temp += temp1;
			temp += fterm;
		}
		else
		{
			STRING temp1;
			temp1.resize(4);
			int k=0;
			Hash::packInt(temp1,mynodes[i]->byteoffset,k);
//~~~~~
			temp += temp1;
			temp += fterm;
		}
	}
	temp += rterm;
	return temp;
}

void BTNode::unpack( BTreeBuffer & buf, int offset, char fterm, char rterm)
{
	STRING temp= buf.unpack(offset);
	int i=0;

	if(!temp[i])
		throw Except("Invalid BTree input file !!!");

	if(temp[i++]!=fterm)
		throw Except("Invalid BTree input file !!!");

	flag = temp[i++];

	if(temp[i++]!=fterm)
		throw Except("Invalid BTree input file !!!");

	unsigned int size = (unsigned int)temp[i++];

	if(temp[i++]!=fterm)
		throw Except("Invalid BTree input file !!!");

	for(int j=0; j<size; j++)
	{
		mynodes.push_back( new BTNCell( mypapa, BTnextToken(temp,i,fterm)));
		i++;
		int z= Hash::unpackInt(temp,i);
		mynodes[j]->byteoffset = z;

		if(flag)
			mynodes[j]->ObjectID = z;
		else
		{	
			mynodes[j]->reference = new BTNode(mypapa);
			mynodes[j]->reference->unpack(buf,z,fterm,rterm);
		}

		i++;
	}
}

void BTNode::out()
{
	int end = mynodes.size();

	if( flag == 1)
	{
		std::cout << "-=( Node " << label << " )=-:" << std::endl;

		for( int count = 0;count < end;count++)
			std::cout << mynodes[count]->key << "\t\t" << mynodes[count]->ObjectID << std::endl;
	}
	else
	{
		for( int count = 0;count < end;count++)
			mynodes[count]->reference->out();
	}
}

void BTNode::find( STRING key)
{
	for( int point = 0;(point < mynodes.size()) && equalbig2keys( key, mynodes[point]->key);point++)
	{
		if( flag == 1)
		{
			if( equal2keys( key, mynodes[point]->key))
				mypapa->lastfounds.push_back( mynodes[point]->ObjectID);
		}
		else
			mynodes[point]->reference->find( key);
	}


	if( point < mynodes.size() && flag != 1)
			mynodes[point]->reference->find( key);
}
