// Buffer.cpp: implementation of the Buffer class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "mfc.h"
#include "Buffer.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

#include "Buffer.h"
#include "Except.h"
#include "AvailList.h"

STRING char2str( const char * input, int count)
{
	STRING retval;

	for(int go = 0;go < count;go++)
		retval += input[go];

	return retval;
}

Buffer::Buffer( Cache * cache, char type, AvailHandler *availhandler, char fieldterm, char recterm)
{
	this->type = type;
	this->cache = cache;
	this->recterm = recterm;
	this->fieldterm = fieldterm;
	this->availhandler = availhandler;
}

Buffer::~Buffer()
{
}

OFFSET Buffer::pack( STRING mustpacked)
{
	OFFSET retval;

	if( mustpacked.size() > BUFSIZE)
		throw Except( "ERR : Your person object is too big!!");

	FreeSpace bestfit( -1, -1);

	try
	{
		bestfit = availhandler->findbest( mustpacked.size());
	}
	catch(NotFound & e)
	{
		retval = cache->write2end( type, mustpacked);
		availhandler->Add( BUFSIZE-mustpacked.size(), retval+mustpacked.size());

		return retval;
	}

	retval = cache->write( type, bestfit.reference, mustpacked);
	availhandler->Add( bestfit.size - mustpacked.size(), retval + mustpacked.size());

	return retval;
}

STRING Buffer::unpack( OFFSET location)
{
	STRING mustunpacked;
	STRING buffer;

	cache->read( type, location, buffer);

	mustunpacked += buffer[0];
	mustunpacked += buffer[1];
	mustunpacked += buffer[2];
	mustunpacked += buffer[3];

	int allocator=0;

	for(allocator = 4;allocator < buffer.size() && buffer[allocator] != recterm;allocator++)
		mustunpacked += buffer[allocator];

	return mustunpacked;
}

void Buffer::remove( OFFSET location)
{
	STRING buffer;
	cache->read( type, location, buffer);
	int end = buffer.size();

	for(int length = 4;length < end && buffer[length] != recterm;length++);

	if(length == end)
		length--;

	availhandler->Add( length + 1, location);

	buffer.resize(0);
	buffer.resize(length + 1);
	
	cache->write( type, location, buffer);
}

STRING Buffer::directread( OFFSET location)
{
	STRING retval;
	cache->read( type, location, retval);

	return retval;
}

int Buffer::directwrite( STRING mustwrite, OFFSET location)
{
	int retval;

	if(location > -1)
	{
		cache->write( type, location, mustwrite);
		return location;
	}

	FreeSpace bestfit( -1, -1);

	try
	{
		bestfit = availhandler->findbest( mustwrite.size());
	}
	catch(NotFound & e)
	{
		retval = cache->write2end( type, mustwrite);
		availhandler->Add( BUFSIZE-mustwrite.size(), retval+mustwrite.size());

		return retval;
	}

	retval = cache->write( type, bestfit.reference, mustwrite);
	availhandler->Add( bestfit.size - mustwrite.size(), retval + mustwrite.size());

	return retval;
}

