// Buffer.h: interface for the Buffer class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_BUFFER_H__144475B2_971F_40F9_9304_0DA8246655A2__INCLUDED_)
#define AFX_BUFFER_H__144475B2_971F_40F9_9304_0DA8246655A2__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#ifndef __BUFFER_H_
#define __BUFFER_H_

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ INCLUDES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#include <fstream>
#include <string>
#include "AvailList.h"
#include "Cache.h"

class BTree;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ TYPES ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
typedef std::fstream			FSTREAM;
typedef std::string				STRING;
typedef std::streamoff			OFFSET;
typedef std::ios_base::seekdir	DIRECTION;

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ CLASS ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class Buffer
{
public:
	Buffer( Cache * cache, char type, AvailHandler * availhandler, char fieldterm = '|', char recterm = '\n');
	~Buffer();

	OFFSET			pack( STRING mustpacked);
	virtual STRING	unpack( OFFSET location);
	void			remove( OFFSET location);

	STRING			directread( OFFSET location);
	int				directwrite( STRING mustwrite, OFFSET location = -1);

	char			getfieldterm(){return fieldterm;}
	char			getrecordterm(){return recterm;}

protected:

	char			type;
	Cache			*cache;
	AvailHandler	*availhandler;
	char			recterm;
	char			fieldterm;
};

class HashBuffer : public Buffer
{
public:
	HashBuffer( Cache * cache, char type, AvailHandler * availhandler,char fieldterm = '|', char recterm = '\n');

	void	remove( OFFSET location, int size=8);
	void	deadbucket( OFFSET location);
private:
	int		spacesize, bucketsize;
};


class BTreeBuffer : public Buffer
{
public:
	BTreeBuffer( Cache * cache, char type, AvailHandler * availhandler, char fieldterm = '|', char recterm = '\n');
	
	void	remove( OFFSET location, int size);
	STRING	unpack( OFFSET location);
};

#endif

#endif // !defined(AFX_BUFFER_H__144475B2_971F_40F9_9304_0DA8246655A2__INCLUDED_)
